/* 
   Copyright (C) 2010, 2011, 2012 German A. Arias <german@xelalug.org>

   This file is part of Gemas application

   Gemas is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 3 of the License, or (at your option) any later version.
 
   This application is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
 
   You should have received a copy of the GNU General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
*/

#import "GemasEditorView.h"
#import <Foundation/Foundation.h>
#import "GemasDocument.h"
#import <HighlighterKit/HighlighterKit.h>

@interface GemasEditorView (Private)

- (BOOL) isGSmarkupIndent: (NSString *)string;
- (BOOL) isGSmarkupBackIndent: (NSString *)string;

@end

@implementation GemasEditorView (Private)

- (BOOL) isGSmarkupIndent: (NSString *)string
{
  if ([type isEqualToString: @"GSmarkup"])
    {
      if ([string hasPrefix: @"<"] &&
          [string hasSuffix: @">"] &&
          ![string hasPrefix: @"<!"] &&
          ![string hasPrefix: @"<?"] &&
          ![string hasPrefix: @"</"] &&
          ![string hasSuffix: @"/>"] &&
          ![string hasSuffix: @"</label>"] &&
          ![string hasSuffix: @"</textField>"] &&
          ![string hasSuffix: @"</secureTextField>"])
        {
          return YES;
        }
      else
        {
          return NO;
        }
    }
  else
    {
      return NO;
    }
}

- (BOOL) isGSmarkupBackIndent: (NSString *)string
{
  return NO;
}

@end

@implementation GemasEditorView

- (void) dealloc
{
  TEST_RELEASE(highlighter);
  TEST_RELEASE(openCharacters);
  TEST_RELEASE(closeCharacters);
  TEST_RELEASE(indentCharacters);
  TEST_RELEASE(backIndentCharacters);
  TEST_RELEASE(nonEnglishCharacters);
  TEST_RELEASE(type);
  [super dealloc];
}

- (void) awakeFromNib
{
  openCharacters = [NSCharacterSet characterSetWithCharactersInString: @"[{("];
  [openCharacters retain];
  closeCharacters = [NSCharacterSet characterSetWithCharactersInString: @"]})"];
  [closeCharacters retain];
  indentCharacters = [NSCharacterSet characterSetWithCharactersInString: @"{"];
  [indentCharacters retain];
  backIndentCharacters = [NSCharacterSet characterSetWithCharactersInString: @"}"];
  [backIndentCharacters retain];
  
  //Dictionary for foreing characters
  NSBundle *bundle = [NSBundle mainBundle];
  nonEnglishCharacters = [NSDictionary dictionaryWithContentsOfFile:
                           [bundle pathForResource: @"nonEnglishCharacters" ofType: @"plist"]];
  [nonEnglishCharacters retain];                         
}

- (void) drawRect: (NSRect) frame
{
  NSLayoutManager *layoutManager;
  NSTextContainer *textContainer;
  NSRange drawnRange;
  
  layoutManager = [self layoutManager];
  textContainer = [self textContainer];
  drawnRange = [layoutManager glyphRangeForBoundingRect: frame
                              inTextContainer: textContainer];
  
  [highlighter highlightRange: drawnRange];
  [super drawRect: frame];
}


//Insert the correponding indentation when the user press Tab key
- (void) insertTab: (id)sender
{
  switch ([[NSUserDefaults standardUserDefaults]
            integerForKey: @"Indentation"])
    {
    case 0:  // 2 spaces
      [super insertText: @"  "];
      break;
    case 1:  // 4 spaces
      [super insertText: @"    "];
      break;
    }
}

//If the new line need indentation, add it
- (void) insertNewline: (id)sender
{
  NSString *spaceToInsert, *previousLine = nil;
  NSString *lastLine;
  NSRange textRange;
  NSArray *lines;
  [super insertNewline: sender];
  textRange = [self selectedRange];
  lines = [[[self string] substringFromRange: NSMakeRange(0, textRange.location)]
                     componentsSeparatedByString: @"\n"];

  if ([lines count] >= 3)
    {
      previousLine = [lines objectAtIndex: [lines count] - 3];
      previousLine = [previousLine stringByReplacingOccurrencesOfString:
                                     @" " withString: @""];
    }

  lastLine = [lines objectAtIndex: [lines count] - 2];
  
  if ([[NSCharacterSet whitespaceAndNewlineCharacterSet] isSupersetOfSet: [NSCharacterSet characterSetWithCharactersInString: lastLine]])
    {
      spaceToInsert = [NSString stringWithString: lastLine];
    }
  else
    {
      unichar lastChar;
      NSInteger tab = [[NSUserDefaults standardUserDefaults]
                        integerForKey: @"Indentation"];
      NSString *checkLine = [lastLine stringByReplacingOccurrencesOfString: @" " withString: @""];
      spaceToInsert = [lastLine stringByDeletingSuffix: [lastLine stringByTrimmingLeadSpaces]];
      
      lastChar = [[lastLine stringByTrimmingTailSpaces] characterAtIndex: [[lastLine stringByTrimmingTailSpaces] length] - 1];
      
      if ([type isEqualToString: @"ChangeLog"])
        {
          if ([[lastLine stringByTrimmingTailSpaces] hasSuffix: @">"])
            {
              spaceToInsert = [NSString stringWithString: @"        "];
            }
        }
      else if ((([indentCharacters characterIsMember: lastChar]) &&
               (![previousLine hasPrefix: @"switch("])) ||
               ([checkLine hasPrefix: @"if("]) ||
               ([checkLine hasPrefix: @"else"]) ||
               ([checkLine hasPrefix: @"while("] && ![previousLine hasPrefix: @"}"]) ||
               ([checkLine hasPrefix: @"for("]) ||
               ([checkLine isEqualToString: @"do"]) ||
               ([checkLine isEqualToString: @"default:"]) ||
               ([checkLine hasPrefix: @"switch("]) ||
               ([checkLine hasPrefix: @"case"]) ||
               ([self isGSmarkupIndent: checkLine]))
        {
          if (tab == 0)
            {
              spaceToInsert = [spaceToInsert stringByAppendingString: @"  "];
            }
          
          if (tab == 1)
            {
              spaceToInsert = [spaceToInsert stringByAppendingString: @"    "];
            }
        }
      else
        {
          if (([backIndentCharacters characterIsMember: lastChar]) ||
              ([checkLine hasSuffix: @"break;"]) ||
              ([self isGSmarkupBackIndent: checkLine]))
            {
              if ((tab == 0) && ([spaceToInsert length] >= 2))
                {
                  spaceToInsert = [spaceToInsert stringByDeletingPrefix: @"  "];
                }
              
              if ((tab == 1) && ([spaceToInsert length] >= 4))
                {
                  spaceToInsert = [spaceToInsert stringByDeletingPrefix: @"    "];
                }
            }
        }
    }

  [super insertText: spaceToInsert];
}

- (void) insertText: (id)string
{
  if ([string isKindOfClass: [NSString class]])
    {
      if ([type isEqualToString: @"Strings"] || [type isEqualToString: @"Plist"])
        {
          if ([[nonEnglishCharacters allKeys] containsObject: string])
            {
              [super insertText: [nonEnglishCharacters objectForKey: string]];
            }
          else if ([string length] > 1)
            {
              int x;
              NSString *character;
              for (x = 0; x < [string length]; x++)
                {
                  character = [string substringFromRange: NSMakeRange(x, 1)];
                  
                  if ([[nonEnglishCharacters allKeys] containsObject: character])
                    {
                      [super insertText: [nonEnglishCharacters objectForKey: character]];
                    }
                  else
                    {
                      [super insertText: character];
                    }
                }
            }
          else
            {
              [super insertText: string];
            }
        }
      else if (([backIndentCharacters characterIsMember: [string characterAtIndex: 0]]) ||
               ([string isEqualToString: @";"]) ||
               ([string isEqualToString: @"c"]) ||
               ([type isEqualToString: @"GSmarkup"] && [string isEqualToString: @"/"]))
        {
          BOOL gsmarkup = NO;
          BOOL backIndent = NO;
          BOOL backIndentCase = NO;
          NSString *prevLine, *lastLine;
          NSRange textRange = [self selectedRange];
          NSArray *lines = [[[self string] substringFromRange: NSMakeRange(0, textRange.location)]
                             componentsSeparatedByString: @"\n"];
                             
          if ([lines count] >= 2)
            {
              prevLine = [lines objectAtIndex: [lines count] - 2];
            }
          else
            {
              prevLine = [NSString stringWithString: @""];
            }
          
          if ([lines count] >= 1)
            {
              lastLine = [lines objectAtIndex: [lines count] - 1];
            }
          else
            {
              lastLine = [NSString stringWithString: @""];
            }
          
          int length;
          NSString *stringToReplace;
          NSCharacterSet *spaces = [NSCharacterSet whitespaceCharacterSet];
          NSCharacterSet *charsLine = [NSCharacterSet characterSetWithCharactersInString: lastLine];
          NSInteger tab = [[NSUserDefaults standardUserDefaults]
                            integerForKey: @"Indentation"];


          if ([[lastLine stringByTrimmingLeadSpaces] isEqualToString: @"<"] && 
              [type isEqualToString: @"GSmarkup"])
            {
              gsmarkup = YES;
            }

          if ([[prevLine stringByTrimmingTailSpaces] hasSuffix: @"}"] &&
              [[lastLine stringByTrimmingLeadSpaces] isEqualToString: @"break"] &&
              [string isEqualToString: @";"])
            {
              backIndent = YES;
            }
          
          if ([[prevLine stringByTrimmingLeadSpaces] hasPrefix: @"case "] &&
              [string isEqualToString: @"c"])
            {
              backIndentCase = YES;
            }
            
          if ((([lastLine length] > 0) && ([spaces isSupersetOfSet: charsLine])) || 
              gsmarkup ||
              backIndent)
            {
              stringToReplace = [NSString stringWithString: lastLine];
            }
          else
            {
              stringToReplace = [NSString stringWithString: @""];
            }
          
          length = [stringToReplace length];

          //Check if is a gsmarkup character
          if (gsmarkup)
            {
              //GSmarkup backIndent
              if ((tab == 0) && ([stringToReplace length] >= 3))
                {
                  stringToReplace = [stringToReplace stringByDeletingPrefix: @"  "];
                }
              
              if ((tab == 1) && ([stringToReplace length] >= 5))
                {
                  stringToReplace = [stringToReplace stringByDeletingPrefix: @"    "];
                }
            }
          else if (backIndent)
            {
              if (tab == 0)
                {
                  stringToReplace = [stringToReplace stringByReplacingString: @"break"
                                                                  withString: @"  break"];
                }
              
              if (tab == 1)
                {
                  stringToReplace = [stringToReplace stringByReplacingString: @"break"
                                                                  withString: @"    break"];
                }
            }
          else if (backIndentCase)
            {
              if ((tab == 0) && ([stringToReplace length] >= 2))
                {
                  stringToReplace = [stringToReplace stringByDeletingPrefix: @"  "];
                }
              
              if ((tab == 1) && ([stringToReplace length] >= 4))
                {
                  stringToReplace = [stringToReplace stringByDeletingPrefix: @"    "];
                }
            }
          else if (![string isEqualToString: @";"] && ![string isEqualToString: @"c"])
            {
              if (![[prevLine stringByTrimmingTailSpaces] hasSuffix: @"break;"])
                {
                  if ((tab == 0) && ([stringToReplace length] >= 2))
                    {
                      stringToReplace = [stringToReplace stringByDeletingPrefix: @"  "];
                    }
                  
                  if ((tab == 1) && ([stringToReplace length] >= 4))
                    {
                      stringToReplace = [stringToReplace stringByDeletingPrefix: @"    "];
                    }
                }
            }
          
          [self replaceCharactersInRange: NSMakeRange(textRange.location - length, length) withString: stringToReplace];
          [super insertText: string];
        }
      else
        {
          [super insertText: string];
        }
    }
  else
    {
      [super insertText: string];
    }
}

/*Select the range of parentheses when the user do a double 
  click obove one*/
- (void) mouseDown: (NSEvent *)event 
{
  NSRange textRange;
  NSString *selection;
  [super mouseDown: event];
  textRange = [self selectedRange];
  selection = [[self string] substringFromRange: textRange];

  if ([selection length] == 1)
    {
      //Open character {[(
      if ([openCharacters characterIsMember: [selection characterAtIndex: 0]])
        {
          NSString *search = nil;
          NSRange firstClose, firstOpen;
          NSString *contextA, *contextB;
          NSString *context = [[self string] substringFromRange: NSMakeRange(textRange.location + 1, [[self string] length] - textRange.location - 1)];

          if ([selection isEqualToString: @"{"])
            {
              search = [NSString stringWithString: @"}"];
            }

          if ([selection isEqualToString: @"["])
            {
              search = [NSString stringWithString: @"]"];
            }

          if ([selection isEqualToString: @"("])
            {
              search = [NSString stringWithString: @")"];
            }

          firstOpen = [context rangeOfString: selection];
          firstClose = [context rangeOfString: search];

          firstOpen = NSMakeRange(firstOpen.location + 1, firstOpen.length);
          firstClose = NSMakeRange(firstClose.location + 1, firstClose.length);

          while ((firstOpen.location < firstClose.location) && (firstOpen.length != 0) && (firstClose.length != 0))
            {
              contextA = [[self string] substringFromRange: NSMakeRange(textRange.location + firstOpen.location + 1, [[self string] length] - firstOpen.location - textRange.location - 1)];
              contextB = [[self string] substringFromRange: NSMakeRange(textRange.location + firstClose.location + 1, [[self string] length] - firstClose.location - textRange.location - 1)];
              
              firstOpen = NSMakeRange([[self string] length] - [contextA length] + [contextA rangeOfString: selection].location - textRange.location, [contextA rangeOfString: selection].length);
              firstClose = NSMakeRange([[self string] length] - [contextB length] + [contextB rangeOfString: search].location - textRange.location, [contextB rangeOfString: search].length);
            }

          [self setSelectedRange: NSMakeRange(textRange.location, firstClose.location + 1)];
        }
      
      //Close character }])
      if ([closeCharacters characterIsMember: [selection characterAtIndex: 0]])
        {
          NSString *search = nil;
          NSRange firstClose, firstOpen;
          NSString *contextA, *contextB;
          NSString *context = [[self string] substringFromRange: NSMakeRange(0, textRange.location)];

          if ([selection isEqualToString: @"}"])
            {
              search = [NSString stringWithString: @"{"];
            }

          if ([selection isEqualToString: @"]"])
            {
              search = [NSString stringWithString: @"["];
            }

          if ([selection isEqualToString: @")"])
            {
              search = [NSString stringWithString: @"("];
            }

          firstOpen = [context rangeOfString: search options: NSBackwardsSearch];
          firstClose = [context rangeOfString: selection options: NSBackwardsSearch];

          while ((firstOpen.location < firstClose.location) && (firstClose.length != 0))
            {
              contextA = [[self string] substringFromRange: NSMakeRange(0, firstOpen.location)];
              contextB = [[self string] substringFromRange: NSMakeRange(0, firstClose.location)];
              
              firstOpen = [contextA rangeOfString: search options: NSBackwardsSearch];
              firstClose = [contextB rangeOfString: selection options: NSBackwardsSearch];
              
            }
          
          [self setSelectedRange: NSMakeRange(firstOpen.location, textRange.location - firstOpen.location + 1)];
        }
    }
}

//Creates a new syntax highlighter for the specified file type
- (void) createSyntaxHighlighterForFileType: (NSString *) fileType
{
  NSString *name = [[editorDocument fileName] lastPathComponent];
  type = [NSString stringWithString: fileType];
  
  if ([type isEqualToString: @"GNUmakefile"] && ([name length] > 0))
    {
      if (![name hasPrefix: @"GNUmakefile"] &&
          ![name isEqualToString: @"ChangeLog"] &&
          ![name isEqualToString: @"Changelog"])
        {
          type = [NSString stringWithString: @"Generic"];
        }
      else if ([name isEqualToString: @"ChangeLog"] ||
               [name isEqualToString: @"Changelog"])
        {
          type = [NSString stringWithString: @"ChangeLog"];
        }
    }
  
  ASSIGN (highlighter, [[[HKSyntaxHighlighter alloc]
                          initWithHighlighterType: type
                          textStorage: [self textStorage]
                          defaultTextColor: [editorDocument textColor]]
                          autorelease]);
                          
  [type retain];

  if ([type isEqualToString: @"C"])
    {
      [[self window] setMiniwindowImage: [NSImage imageNamed: @"FileIcon_.c"]];
    }

  if ([type isEqualToString: @"C-Header"])
    {
      [[self window] setMiniwindowImage: [NSImage imageNamed: @"FileIcon_.h"]];
    }

  if ([type isEqualToString: @"C++"])
    {
      [[self window] setMiniwindowImage: [NSImage imageNamed: @"FileIcon_.cc"]];
    }

  if ([type isEqualToString: @"ObjC"])
    {
      [[self window] setMiniwindowImage: [NSImage imageNamed: @"FileIcon_.m"]];
    }

  if ([type isEqualToString: @"ObjC++"])
    {
      [[self window] setMiniwindowImage: [NSImage imageNamed: @"FileIcon_.mm"]];
    }

  if ([type isEqualToString: @"Plist"])
    {
      [[self window] setMiniwindowImage: [NSImage imageNamed: @"FileIcon_.plist"]];
    }

  if ([type isEqualToString: @"Strings"])
    {
      [[self window] setMiniwindowImage: [NSImage imageNamed: @"FileIcon_.strings"]];
    }

  if ([type isEqualToString: @"GSmarkup"])
    {
      [[self window] setMiniwindowImage: [NSImage imageNamed: @"FileIcon_.gsmarkup"]];
    }

  if ([type isEqualToString: @"GNUmakefile"])
    {
      [[self window] setMiniwindowImage: [NSImage imageNamed: @"FileIcon_makefile"]];
    }
  
  if ([type isEqualToString: @"Generic"] || [type isEqualToString: @"ChangeLog"])
    {
      [[self window] setMiniwindowImage: [NSImage imageNamed: @"common_Unknown"]];
    }
}

@end
